from typing import Literal, TypedDict, assert_never, get_args

type BasicCardColor = Literal["red", "green", "blue", "yellow"]
basic_card_colors: tuple[BasicCardColor, ...] = get_args(BasicCardColor.__value__)


class NumberCard(TypedDict):
    id: str
    color: BasicCardColor
    type: Literal["number"]
    number: int


type FunctionCardEffect = Literal["skip", "+2", "reverse"]
function_card_effects: tuple[FunctionCardEffect, ...] = get_args(
    FunctionCardEffect.__value__
)


class FunctionCard(TypedDict):
    id: str
    color: BasicCardColor
    type: Literal["function"]
    effect: FunctionCardEffect


type WildCardEffect = Literal["+4", "color"]
wild_card_effects: tuple[WildCardEffect, ...] = get_args(WildCardEffect.__value__)
type WildCardColor = Literal["black"]
wild_card_color: WildCardColor = "black"


class WildCard(TypedDict):
    id: str
    color: WildCardColor
    type: Literal["wild"]
    effect: WildCardEffect


type Card = NumberCard | FunctionCard | WildCard
type Deck = list[Card]


def format_card(card: Card) -> str:
    if card["type"] == "number":
        return card["color"].title() + "_" + str(card["number"])
    elif card["type"] == "function":
        return card["color"].title() + "_" + card["effect"].title()
    elif card["type"] == "wild":
        return card["effect"].title()
    else:
        assert_never(card["type"])



from abc import abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from inspect import get_annotations
from typing import (
    Annotated,
    Any,
    NamedTuple,
    Protocol,
    TypedDict,
    cast,
    get_origin,
    override,
    runtime_checkable,
)

from tuno.server.exceptions import ApiException
from tuno.shared.constraints import (
    DEFAULT_BOT_COUNT,
    DEFAULT_BOT_PLAY_DELAY_SECONDS,
    DEFAULT_INITIAL_HAND_SIZE,
    DEFAULT_PLAYER_CAPACITY,
    MAX_BOT_COUNT,
    MAX_BOT_PLAY_DELAY_SECONDS,
    MAX_INITIAL_HAND_SIZE,
    MAX_PLAYER_CAPACITY,
    MIN_BOT_COUNT,
    MIN_BOT_PLAY_DELAY_SECONDS,
    MIN_INITIAL_HAND_SIZE,
    MIN_PLAYER_CAPACITY,
)


@runtime_checkable
class RuleValidator(Protocol):

    rule_name: str

    def __init__(self, rule_name: str) -> None:
        super().__init__()
        self.rule_name = rule_name

    @abstractmethod
    def validate(self, value: object) -> None: ...


class RuleValidationException(ApiException):

    def __init__(self, message: str) -> None:
        super().__init__(400, message)


class RangeRuleValidator(RuleValidator):

    min: float
    max: float

    @property
    @abstractmethod
    def _type(self) -> type[int | float]: ...

    def __init__(self, rule_name: str, min: float, max: float) -> None:
        super().__init__(rule_name)
        self.min = min
        self.max = max

    def validate(self, value: object) -> None:
        if not isinstance(value, self._type):
            raise RuleValidationException(
                f"{self.rule_name} must be of type {self._type.__name__}, "
                f"got: {type(value).__name__}"
            )
        assert isinstance(value, (int, float))
        if not self.min <= value <= self.max:
            raise RuleValidationException(
                f"{self.rule_name} must be in range "
                f"[{self.min}, {self.max}], got: {value}"
            )


class IntRangeRuleValidator(RangeRuleValidator):

    @property
    @override
    def _type(self) -> type[int]:
        return int


class FloatRangeRuleValidator(RangeRuleValidator):

    @property
    @override
    def _type(self) -> type[float]:
        return float


class RuleMetadataWithoutType(NamedTuple):
    default: object
    hint: str
    validator: RuleValidator | None


class GameRules(TypedDict):

    player_capacity: Annotated[
        int,
        RuleMetadataWithoutType(
            default=DEFAULT_PLAYER_CAPACITY,
            hint=f"{MIN_PLAYER_CAPACITY}~{MAX_PLAYER_CAPACITY}",
            validator=IntRangeRuleValidator(
                "player_capacity",
                MIN_PLAYER_CAPACITY,
                MAX_PLAYER_CAPACITY,
            ),
        ),
    ]

    shuffle_players: Annotated[
        bool,
        RuleMetadataWithoutType(
            default=True,
            hint="shuffle players before starting",
            validator=None,
        ),
    ]

    initial_hand_size: Annotated[
        int,
        RuleMetadataWithoutType(
            default=DEFAULT_INITIAL_HAND_SIZE,
            hint=f"{MIN_INITIAL_HAND_SIZE}~{MAX_INITIAL_HAND_SIZE}",
            validator=IntRangeRuleValidator(
                "initial_hand_size",
                MIN_INITIAL_HAND_SIZE,
                MAX_INITIAL_HAND_SIZE,
            ),
        ),
    ]

    any_last_play: Annotated[
        bool,
        RuleMetadataWithoutType(
            default=True,
            hint="allow non-number card as last play",
            validator=None,
        ),
    ]

    bot_count: Annotated[
        int,
        RuleMetadataWithoutType(
            default=DEFAULT_BOT_COUNT,
            hint=f"{MIN_BOT_COUNT}~{MAX_BOT_COUNT}",
            validator=IntRangeRuleValidator(
                "bot_count",
                MIN_BOT_COUNT,
                MAX_BOT_COUNT,
            ),
        ),
    ]

    bot_play_delay: Annotated[
        float,
        RuleMetadataWithoutType(
            default=DEFAULT_BOT_PLAY_DELAY_SECONDS,
            hint=f"{MIN_BOT_PLAY_DELAY_SECONDS}s~{MAX_BOT_PLAY_DELAY_SECONDS}s",
            validator=FloatRangeRuleValidator(
                "bot_play_delay",
                MIN_BOT_PLAY_DELAY_SECONDS,
                MAX_BOT_PLAY_DELAY_SECONDS,
            ),
        ),
    ]


@dataclass
class RuleMetadata:
    type: type
    default: object
    hint: str
    validator: RuleValidator | None


@runtime_checkable
class AnnotatedType(Protocol):
    """A partial type for interacting with builtin _AnnotatedAlias, which is not exported."""

    __origin__: type
    __metadata__: tuple[object, ...]


def __parse_rule_metadata(type_annotation: object) -> RuleMetadata:

    if get_origin(type_annotation) is not Annotated:
        raise RuleValidationException(f"Invalid rule annotation: {type_annotation}")
    assert isinstance(type_annotation, AnnotatedType)

    type_metadata = type_annotation.__metadata__
    if (len(type_metadata) != 1) or (
        not isinstance(
            rule_metadata_without_type := type_metadata[0],
            RuleMetadataWithoutType,
        )
    ):
        raise RuntimeError(f"Invalid metadata: {type_metadata!r}")

    return RuleMetadata(
        type=type_annotation.__origin__,
        default=rule_metadata_without_type.default,
        hint=rule_metadata_without_type.hint,
        validator=rule_metadata_without_type.validator,
    )


rule_metadata_map: Mapping[str, RuleMetadata] = {
    key: __parse_rule_metadata(type_annotation)
    for key, type_annotation in get_annotations(GameRules).items()
}


def create_game_rules() -> GameRules:
    return cast(
        GameRules,
        {
            key: cast(Any, metadata.default)
            for key, metadata in rule_metadata_map.items()
        },
    )


def check_rule_update(key: str, value: object) -> None:

    rule_metadata = rule_metadata_map.get(key, None)
    if not rule_metadata:
        raise RuleValidationException(f"Unknown rule: {key}")

    if not isinstance(value, rule_metadata.type):
        raise RuleValidationException(
            f"Invalid type for rule `{key}`: "
            f"expected {rule_metadata.type}, got {type(value)}"
        )

    if rule_metadata.validator:
        rule_metadata.validator.validate(value)



from collections.abc import Generator, Iterable, Sequence
from contextlib import contextmanager
from queue import Queue
from random import choice
from threading import RLock
from time import monotonic

from tuno.server.config import PLAYER_MESSAGE_QUEUE_SIZE
from tuno.server.exceptions import CardIdsNotFoundException
from tuno.server.utils.Logger import Logger
from tuno.shared.check_play import InvalidPlayException, check_play
from tuno.shared.constraints import DEFAULT_BOT_PLAY_DELAY_SECONDS
from tuno.shared.deck import BasicCardColor, Card, Deck, basic_card_colors
from tuno.shared.LazyTimer import LazyTimer
from tuno.shared.rules import GameRules
from tuno.shared.sse_events import CardsEvent, ServerSentEvent


class Player:

    is_bot: bool
    name: str
    cards: Deck
    last_result: int
    message_queue: Queue[ServerSentEvent]
    lock: RLock
    connected: bool  # set by game watcher
    subscription_token: str
    bot_play_timer: LazyTimer | None
    __logger: Logger
    __last_pending_timestamp: float | None
    __last_sent_timestamp: float | None

    def __init__(self, name: str, *, is_bot: bool) -> None:

        self.is_bot = is_bot
        self.name = name
        self.cards = []
        self.last_result = -1
        self.message_queue = Queue(PLAYER_MESSAGE_QUEUE_SIZE)
        self.lock = RLock()
        self.connected = False
        self.subscription_token = ""
        self.bot_play_timer = (
            LazyTimer(DEFAULT_BOT_PLAY_DELAY_SECONDS) if is_bot else None
        )
        self.__logger = Logger(f"{__name__}#{name}")
        self.__last_pending_timestamp = None
        self.__last_sent_timestamp = None

        self.__logger.debug(f"player#{name} created")

    @property
    def last_pending_timestamp(self) -> float | None:
        return self.__last_pending_timestamp

    @property
    def last_sent_timestamp(self) -> float | None:
        return self.__last_sent_timestamp

    @contextmanager
    def message_context(self) -> Generator[None]:
        with self.lock:
            self.__last_pending_timestamp = monotonic()
            yield
            self.__last_pending_timestamp = None
            self.__last_sent_timestamp = monotonic()

    def get_cards_event(self) -> CardsEvent:
        return CardsEvent(self.cards)

    def give_out_cards(self, card_ids: Sequence[str]) -> Deck:

        if len(card_ids) == 0:
            return []

        cards_ids_remaining = set(card_ids)
        cards_left: Deck = []
        cards_out: Deck = []

        with self.lock:

            for card in self.cards:
                if card["id"] in cards_ids_remaining:
                    cards_ids_remaining.remove(card["id"])
                    cards_out.append(card)
                else:
                    cards_left.append(card)

            if len(cards_ids_remaining) > 0:
                raise CardIdsNotFoundException(cards_ids_remaining)

            self.cards = cards_left

        return cards_out

    def bot_play(
        self,
        *,
        lead_color: BasicCardColor,
        lead_card: Card,
        skip_counter: int,
        rules: GameRules,
    ) -> tuple[list[str], BasicCardColor | None]:
        with self.lock:

            candidates: list[tuple[list[str], BasicCardColor | None]] = []
            for card in self.cards:
                colors: Iterable[BasicCardColor | None] = (
                    basic_card_colors if card["type"] == "wild" else [None]
                )
                for color in colors:
                    try:
                        check_play(
                            [card],
                            color,
                            lead_color=lead_color,
                            lead_card=lead_card,
                            skip_counter=skip_counter,
                            rules=rules,
                        )
                    except InvalidPlayException:
                        continue
                    else:
                        candidates.append(([card["id"]], color))

            if len(candidates) == 0:
                return ([], None)  # pass
            else:
                return choice(candidates)



from time import monotonic


class LazyTimer:

    timeout: float

    __start_time: float | None = None

    def __init__(self, timeout: float) -> None:
        self.timeout = timeout

    @property
    def started(self) -> bool:
        return self.__start_time is not None

    @started.setter
    def started(self, value: bool) -> None:
        if value:
            self.__start_time = monotonic()
        else:
            self.__start_time = None

    @property
    def time_up(self) -> bool:
        start_time = self.__start_time
        if start_time is None:
            return False
        else:
            now = monotonic()
            return now - start_time >= self.timeout



import re
from typing import Final

PLAYER_NAME_PATTERN: Final = re.compile(r"^[A-Za-z0-9_-]{1,20}$")

MIN_PLAYER_CAPACITY: Final = 2
MAX_PLAYER_CAPACITY: Final = 20
DEFAULT_PLAYER_CAPACITY: Final = 8
assert MIN_PLAYER_CAPACITY <= DEFAULT_PLAYER_CAPACITY <= MAX_PLAYER_CAPACITY

MIN_INITIAL_HAND_SIZE: Final = 2
MAX_INITIAL_HAND_SIZE: Final = 16
DEFAULT_INITIAL_HAND_SIZE: Final = 7
assert MIN_INITIAL_HAND_SIZE <= DEFAULT_INITIAL_HAND_SIZE <= MAX_INITIAL_HAND_SIZE

MIN_BOT_COUNT: Final = 0
MAX_BOT_COUNT: Final = 10
DEFAULT_BOT_COUNT: Final = 0
assert MIN_BOT_COUNT <= DEFAULT_BOT_COUNT <= MAX_BOT_COUNT

MIN_BOT_PLAY_DELAY_SECONDS: Final = 0.0
MAX_BOT_PLAY_DELAY_SECONDS: Final = 10.0
DEFAULT_BOT_PLAY_DELAY_SECONDS: Final = 2.5
assert (
    MIN_BOT_PLAY_DELAY_SECONDS
    <= DEFAULT_BOT_PLAY_DELAY_SECONDS
    <= MAX_BOT_PLAY_DELAY_SECONDS
)
